@@ -19,10 +19,12 @@ use constant BROKEN_SYMLINK   => bless {}, "A::BROKEN::SYMLINK";
 use constant CIRCULAR_SYMLINK => bless {}, "A::CIRCULAR::SYMLINK";
 
 # we're going to use carp but the errors should come from outside of our package.
-use Carp qw(carp confess);
-$Carp::Internal{__PACKAGE__}++;
-$Carp::Internal{'Overload::FileCheck'}++;
+use Carp qw(carp confess croak);
 
+BEGIN {
+    $Carp::Internal{ (__PACKAGE__) }++;
+    $Carp::Internal{'Overload::FileCheck'}++;
+}
 use Cwd                        ();
 use IO::File                   ();
 use Test::MockFile::FileHandle ();
@@ -45,11 +47,11 @@ files without touching the file system.
 
 =head1 VERSION
 
-Version 0.028
+Version 0.029
 
 =cut
 
-our $VERSION = '0.028';
+our $VERSION = '0.029';
 
 our %files_being_mocked;
 
@@ -71,8 +73,8 @@ Intercepts file system calls for specific files so unit testing can
 take place without any files being altered on disk.
 
 This is useful for L<small
-tests|https://en.wikipedia.org/wiki/Google_Test#Small_Tests_(Unit_Tests)>
-where file interaction is discouraged.
+tests|https://testing.googleblog.com/2010/12/test-sizes.html> where
+file interaction is discouraged.
 
 A strict mode is even provided (and turned on by default) which can
 throw a die when files are accessed during your tests!
@@ -81,41 +83,50 @@ throw a die when files are accessed during your tests!
     use Module::I::Dont::Want::To::Alter;
 
     # strict mode by default
-    use Test::MockFile;
+    use Test::MockFile ();
 
     # non-strict mode
     use Test::MockFile qw< nostrict >;
 
     # Be sure to assign the output of mocks, they disappear when they go out of scope
-    my $mock_file = Test::MockFile->file("/foo/bar", "contents\ngo\nhere");
-    open my $fh, '<', '/foo/bar' or die; # Does not actually open the file on disk
-    say 'ok' if -e $fh;
+    my $foobar = Test::MockFile->file( "/foo/bar", "contents\ngo\nhere" );
+    open my $fh, '<', '/foo/bar' or die;    # Does not actually open the file on disk
+    say '/foo/bar exists' if -e $fh;
     close $fh;
-    say 'ok' if -f '/foo/bar';
+
+    say '/foo/bar is a file' if -f '/foo/bar';
     say '/foo/bar is THIS BIG: ' . -s '/foo/bar';
 
-    my $missing_mocked_file = Test::MockFile->file('/foo/baz'); # File starts out missing
-    my $opened              = open my $baz_fh, '<', '/foo/baz'; # File reports as missing so fails
-    say 'ok' if !-e '/foo/baz';
+    my $foobaz = Test::MockFile->file('/foo/baz');    # File starts out missing
+    my $opened = open my $baz_fh, '<', '/foo/baz';    # File reports as missing so fails
+    say '/foo/baz does not exist yet' if !-e '/foo/baz';
 
-    open $baz_fh, '>', '/foo/baz' or die; # open for writing
-    print {$baz_fh} "replace contents\n";
+    open $baz_fh, '>', '/foo/baz' or die;             # open for writing
+    print {$baz_fh} "first line\n";
 
-    open $baz_fh, '>>', '/foo/baz' or die; # open for append.
+    open $baz_fh, '>>', '/foo/baz' or die;            # open for append.
     print {$baz_fh} "second line";
     close $baz_fh;
 
-    say $baz->contents;
+    say "Contents of /foo/baz:\n>>" . $foobaz->contents() . '<<';
 
     # Unmock your file.
-    undef $missing_mocked_file;
+    # (same as the variable going out of scope
+    undef $foobaz;
 
     # The file check will now happen on file system now the file is no longer mocked.
-    say 'ok' if !-e '/foo/baz';
+    say '/foo/baz is missing again (no longer mocked)' if !-e '/foo/baz';
 
-    my $quux    = Test::MockFile->file('/foo/bar/quux.txt');
-    my @matches = </foo/bar/*.txt>;       # ( 'quux.txt' )
-    my @matches = glob('/foo/bar/*.txt'); # same as above
+    my $quux    = Test::MockFile->file( '/foo/bar/quux.txt', '' );
+    my @matches = </foo/bar/*.txt>;
+
+    # ( '/foo/bar/quux.txt' )
+    say "Contents of /foo/bar directory: " . join "\n", @matches;
+
+    @matches = glob('/foo/bar/*.txt');
+
+    # same as above
+    say "Contents of /foo/bar directory (using glob()): " . join "\n", @matches;
 
 =head1 IMPORT
 
@@ -155,13 +166,20 @@ Relative paths are not supported:
 
 =cut
 
+use constant STRICT_MODE_DISABLED => 1;
+use constant STRICT_MODE_ENABLED  => 2;
+use constant STRICT_MODE_UNSET    => 4;
+use constant STRICT_MODE_DEFAULT  => STRICT_MODE_ENABLED | STRICT_MODE_UNSET;    # default state when unset by user
+
 our %authorized_strict_mode_packages;
+our $STRICT_MODE_STATUS;
 
 BEGIN {
     %authorized_strict_mode_packages = (
         'DynaLoader' => 1,
         'lib'        => 1,
     );
+    $STRICT_MODE_STATUS = STRICT_MODE_DEFAULT;
 }
 
 # Perl understands barewords are filehandles during compilation and
@@ -203,18 +221,47 @@ sub _upgrade_barewords {
     return @args;
 }
 
+=head2 file_arg_position_for_command
+
+Args: ($command)
+
+Provides a hint with the position of the argument most likely
+holding the file name for the current C<$command> call.
+
+This is used internaly to provide better error messages.
+This can be used when plugging hooks to know what's the filename
+we currently try to access.
+
+=cut
+
+my $_file_arg_post;
+
+sub file_arg_position_for_command {    # can also be used by user hooks
+    my ($command) = @_;
+
+    $_file_arg_post //= {
+        'chmod'    => 2,
+        'chown'    => 2,
+        'lstat'    => 0,
+        'mkdir'    => 0,
+        'open'     => 2,
+        'opendir'  => 1,
+        'readlink' => 0,
+        'rmdir'    => 0,
+        'stat'     => 0,
+        'sysopen'  => 1,
+        'unlink'   => 0,
+    };
+
+    croak("Unknown strict mode violation for $command") unless defined $command && defined $_file_arg_post->{$command};
+
+    return $_file_arg_post->{$command};
+}
+
 sub _strict_mode_violation {
     my ( $command, $at_under_ref ) = @_;
 
-    my $file_arg =
-        $command eq 'open'    ? 2
-      : $command eq 'sysopen' ? 1
-      : $command eq 'opendir' ? 1
-      : $command eq 'stat'    ? 0
-      : $command eq 'lstat'   ? 0
-      : $command eq 'chown'   ? 2
-      : $command eq 'chmod'   ? 2
-      :                         confess("Unknown strict mode violation for $command");
+    return unless $STRICT_MODE_STATUS & STRICT_MODE_ENABLED;
 
     my @stack;
     foreach my $stack_level ( 1 .. 100 ) {
@@ -232,6 +279,9 @@ sub _strict_mode_violation {
         last;
     }
 
+    # check it later so we give priority to authorized_strict_mode_packages
+    my $file_arg = file_arg_position_for_command($command);
+
     if ( $command eq 'open' and scalar @$at_under_ref != 3 ) {
         $file_arg = 1 if scalar @$at_under_ref == 2;
     }
@@ -247,8 +297,20 @@ sub _strict_mode_violation {
 sub import {
     my ( $class, @args ) = @_;
 
-    grep { $_ eq 'nostrict' } @args
-      and clear_file_access_hooks();
+    my $strict_mode = ( grep { $_ eq 'nostrict' } @args ) ? STRICT_MODE_DISABLED : STRICT_MODE_ENABLED;
+
+    if (
+        defined $STRICT_MODE_STATUS
+        && !( $STRICT_MODE_STATUS & STRICT_MODE_UNSET )    # mode is set by user
+        && $STRICT_MODE_STATUS != $strict_mode
+    ) {
+
+        # could consider using authorized_strict_mode_packages for all packages
+        die q[Test::MockFile is imported multiple times with different strict modes (not currently supported) ] . $class;
+    }
+    $STRICT_MODE_STATUS = $strict_mode;
+
+    return;
 }
 
 =head1 SUBROUTINES/METHODS
@@ -282,7 +344,7 @@ sub file {
     _get_file_object($file) and confess("It looks like $file is already being mocked. We don't support double mocking yet.");
 
     my $path = _abs_path_to_file($file);
-    _validate_path($path);
+    _validate_path($_) for $file, $path;
 
     if ( @stats > 1 ) {
         confess(
@@ -405,6 +467,8 @@ sub _validate_path {
     if ( $path =~ m{ ( ^ | / ) \.{2} ( / | $ ) }xms ) {
         confess('Relative paths are not supported');
     }
+
+    return;
 }
 
 =head2 dir
@@ -472,7 +536,7 @@ sub dir {
       and confess("It looks like $dirname is already being mocked. We don't support double mocking yet.");
 
     my $path = _abs_path_to_file($dirname);
-    _validate_path($path);
+    _validate_path($_) for $dirname, $path;
 
     # Cleanup trailing forward slashes
     $path ne '/'
@@ -623,8 +687,8 @@ sub _mock_stat {
     # Find the path, following the symlink if required.
     my $file = _find_file_or_fh( $file_or_fh, $follow_link );
 
-    return [] if $file && $file eq BROKEN_SYMLINK;      # Allow an ELOOP to fall through here.
-    return [] if $file && $file eq CIRCULAR_SYMLINK;    # Allow an ELOOP to fall through here.
+    return [] if defined $file && defined BROKEN_SYMLINK   && $file eq BROKEN_SYMLINK;      # Allow an ELOOP to fall through here.
+    return [] if defined $file && defined CIRCULAR_SYMLINK && $file eq CIRCULAR_SYMLINK;    # Allow an ELOOP to fall through here.
 
     if ( !defined $file or !length $file ) {
         _real_file_access_hook( $type, [$file_or_fh] );
@@ -726,11 +790,22 @@ sub _abs_path_to_file {
 
     defined $path or return;
 
-    # cleanup multiple slashes
-    $path =~ s{//+}{/}xmsg;
+    my $match = 1;
+    while ($match) {
+        $match = 0;
+        $match = 1 if $path =~ s{//+}{/}xmsg;                # cleanup multiple slashes
+        $match = 1 if $path =~ s{/\.$}{/};
+        $match = 1 if $path =~ s{(?:[^/]+)/\.\.(/|$)}{$1};
+        $match = 1 if $path =~ s{/$}{};
+    }
+
+    return q[/] if $path eq q[/..];
 
     return $path if $path =~ m{^/};
 
+    my $cwd = Cwd::getcwd();
+
+    return $cwd if $path eq '.';
     return Cwd::getcwd() . "/$path";
 }
 
@@ -824,7 +899,7 @@ Deprecated. Same as C<path>.
 =cut
 
 sub filename {
-    warn 'filename() is deprecated, use path() instead';
+    carp('filename() is deprecated, use path() instead');
     goto &path;
 }
 
@@ -1168,13 +1243,15 @@ One use might be:
 
 =cut
 
-my @file_access_hooks = ( \&_strict_mode_violation );
+# always use the _strict_mode_violation
+my @_public_access_hooks;
+my @_internal_access_hooks = ( \&_strict_mode_violation );
 
 sub add_file_access_hook {
     my ($code_ref) = @_;
 
     ( $code_ref && ref $code_ref eq 'CODE' ) or confess("add_file_access_hook needs to be passed a code reference.");
-    push @file_access_hooks, $code_ref;
+    push @_public_access_hooks, $code_ref;
 
     return 1;
 }
@@ -1187,7 +1264,7 @@ B<add_file_access_hook>
 =cut
 
 sub clear_file_access_hooks {
-    @file_access_hooks = ();
+    @_public_access_hooks = ();
 
     return 1;
 }
@@ -1197,7 +1274,7 @@ sub clear_file_access_hooks {
 sub _real_file_access_hook {
     my ( $access_type, $at_under_ref ) = @_;
 
-    foreach my $code (@file_access_hooks) {
+    foreach my $code ( @_internal_access_hooks, @_public_access_hooks ) {
         $code->( $access_type, $at_under_ref );
     }
 
@@ -1580,7 +1657,7 @@ BEGIN {
         my $mocked_dir = _get_file_object( $_[0] );
 
         if ( !$mocked_dir ) {
-            _real_file_access_hook( 'readdir', [ \@_ ] );
+            _real_file_access_hook( 'readdir', \@_ );
             goto \&CORE::readdir if _goto_is_available();
             return CORE::readdir( $_[0] );
         }
@@ -1622,7 +1699,7 @@ BEGIN {
         my $mocked_dir = _get_file_object($fh);
 
         if ( !$mocked_dir || !$mocked_dir->{'obj'} ) {
-            _real_file_access_hook( 'telldir', [ \@_ ] );
+            _real_file_access_hook( 'telldir', \@_ );
             goto \&CORE::telldir if _goto_is_available();
             return CORE::telldir($fh);
         }
@@ -1649,7 +1726,7 @@ BEGIN {
         my $mocked_dir = _get_file_object($fh);
 
         if ( !$mocked_dir || !$mocked_dir->{'obj'} ) {
-            _real_file_access_hook( 'rewinddir', [ \@_ ] );
+            _real_file_access_hook( 'rewinddir', \@_ );
             goto \&CORE::rewinddir if _goto_is_available();
             return CORE::rewinddir( $_[0] );
         }
@@ -1677,7 +1754,7 @@ BEGIN {
         my $mocked_dir = _get_file_object($fh);
 
         if ( !$mocked_dir || !$mocked_dir->{'obj'} ) {
-            _real_file_access_hook( 'seekdir', [ \@_ ] );
+            _real_file_access_hook( 'seekdir', \@_ );
             goto \&CORE::seekdir if _goto_is_available();
             return CORE::seekdir( $fh, $goto );
         }
@@ -1704,7 +1781,7 @@ BEGIN {
         my $mocked_dir = _get_file_object($fh);
 
         if ( !$mocked_dir || !$mocked_dir->{'obj'} ) {
-            _real_file_access_hook( 'closedir', [ \@_ ] );
+            _real_file_access_hook( 'closedir', \@_ );
             goto \&CORE::closedir if _goto_is_available();
             return CORE::closedir($fh);
         }
@@ -1739,7 +1816,7 @@ BEGIN {
         my ($file) = @_;
 
         if ( !defined $file ) {
-            warn 'Use of uninitialized value in readlink';
+            carp('Use of uninitialized value in readlink');
             if ( $^O eq 'freebsd' ) {
                 $! = EINVAL;
             }
@@ -1751,7 +1828,7 @@ BEGIN {
 
         my $mock_object = _get_file_object($file);
         if ( !$mock_object ) {
-            _real_file_access_hook( 'readlink', [ \@_ ] );
+            _real_file_access_hook( 'readlink', \@_ );
             goto \&CORE::readlink if _goto_is_available();
             return CORE::readlink($file);
         }
@@ -1780,7 +1857,7 @@ BEGIN {
         my $mock = _get_file_object($file);
 
         if ( !$mock ) {
-            _real_file_access_hook( 'mkdir', [ \@_ ] );
+            _real_file_access_hook( 'mkdir', \@_ );
             goto \&CORE::mkdir if _goto_is_available();
             return CORE::mkdir(@_);
         }
@@ -1816,7 +1893,7 @@ BEGIN {
         my $mock = _get_file_object($file);
 
         if ( !$mock ) {
-            _real_file_access_hook( 'rmdir', [ \@_ ] );
+            _real_file_access_hook( 'rmdir', \@_ );
             goto \&CORE::rmdir if _goto_is_available();
             return CORE::rmdir($file);
         }
